# $Id: Web.pm,v 1.22 2002/08/05 05:53:56 devin2 Exp $

# the web server portion of our program.

use strict;

package Oulibot::Server::Web;

use Socket;
use HTTP::Response;

use POE::Session;
use POE::Preprocessor;
use POE::Component::Server::TCP;
use POE::Filter::HTTPD;
use Data::Dumper;
use Template;

use Oulibot::Util::Conf;
use Oulibot::Util::Web;
#use Util::Link;

# Dumps the request to stderr.
sub DUMP_REQUEST () { 0 }

sub WEBLOG_TYPE () { "webface" }

#------------------------------------------------------------------------------
# A web server.

# Start an HTTPD session.  Note that this handler receives both the
# local bind() address ($my_host) and the public server address
# ($my_ifname).  It uses $my_ifname to build HTML that the outside
# world can see.

sub httpd_session_started {
  my ( $heap,
       $socket, $remote_address, $remote_port,
       $my_name, $my_host, $my_port, $my_ifname, $login, $passwd,$template_root
     ) = @_[HEAP, ARG0..ARG9];

  # TODO: I think $my_host is obsolete.  Maybe it can be removed, and
  # $my_ifname can be used exclusively?

  $heap->{my_host} = $my_host;
  $heap->{my_port} = $my_port;
  $heap->{my_name} = $my_name;
  $heap->{my_inam} = $my_ifname;
  $heap->{login}   = $login;
  $heap->{passwd}  = $passwd;
  $heap->{troot} = $template_root;

  $heap->{remote_addr} = inet_ntoa($remote_address);
  $heap->{remote_port} = $remote_port;

  $heap->{wheel} = new POE::Wheel::ReadWrite
    ( Handle       => $socket,
      Driver       => new POE::Driver::SysRW,
      Filter       => new POE::Filter::HTTPD,
      InputEvent   => 'got_query',
      FlushedEvent => 'got_flush',
      ErrorEvent   => 'got_error',
    );
}

# An HTTPD response has flushed.  Stop the session.
sub httpd_session_flushed {
  delete $_[HEAP]->{wheel};
}

# An HTTPD session received an error.  Stop the session.
sub httpd_session_got_error {
  my ($session, $heap, $operation, $errnum, $errstr) =
    @_[SESSION, HEAP, ARG0, ARG1, ARG2];
  warn( "connection session ", $session->ID,
        " got $operation error $errnum: $errstr\n"
      );
  delete $heap->{wheel};
}

# Process HTTP requests.
sub httpd_session_got_query {
  my ($heap, $request) = @_[HEAP, ARG0];

  ### Log the request.

  # Space-separated list:
  # Remote address (client address)
  # -
  # -
  # [GMT date in brackets: DD/Mon/CCYY:HH:MM:SS -0000]
  # "GET url HTTP/x.y"  <-- in quotes
  # response code
  # response size
  # referer
  # user-agent string

  ### Responded with an error.  Send it directly.

  if ($request->isa("HTTP::Response")) {
    $heap->{wheel}->put($request);
    return;
  }

  ###---------------------------------------------
  ### These requests don't require authentication.
  ###---------------------------------------------

  my $url = $request->url() . '';
  $url .= 'index.html' if not $url or ($url eq '/');

  my ($root,$branch) = $url =~ /\/?(\w+)?\/(.+)/;
  my $content = parse_content($request->content());
  warn(Dumper($content).'-> content');
  ### Root page.

  if ($root eq 'rdf') {

  }
  elsif ($branch eq 'index.html') {
    my $response = HTTP::Response->new(200);
    $response->push_header( 'Content-type', 'text/html' );

    my $source = $root || $content->{source} || 'oulibot';

    my $oulibot = Oulibot->new(table => 'pages',
			       source => $source);

    my $question = $content->{question} || $oulibot->random_past_input;
    my @responses = $oulibot->contextualise($question);

    my $html;
    my $params = {proposition => $question,
		  aphorisms => \@responses};


    $heap->{tt} ||= Template->new({
	INCLUDE_PATH  => $heap->{troot},
	PLUGIN_BASE   => 'Oulibot::Template::Plugin'
	});
    my $tt = $heap->{tt};
    my $template = 'index.html';

    $tt->process($template,$params,\$html) || warn $tt->error();

    $response->content($html);

    $heap->{wheel}->put( $response );
    return;
  }

  ### Deny robots.

  if ($url eq '/robots.txt') {
    my $response = HTTP::Response->new(200);
    $response->push_header( 'Content-type', 'text/plain' );
    $response->content
      ( "User-agent: *\x0d\x0a" .
        "Disallow: /\x0d\x0a"
      );
    $heap->{wheel}->put($response);
    return;
  }

  ### Add a link via the web.  For example, via a javascript bookmark.

  if ($url =~ /^\/add\?(.+?)\s*$/) {
    my $link = $1;

    my $response = HTTP::Response->new(200);
    $response->push_header( 'Content-type', 'text/html' );

    if ($link =~ s/(http:\/\/\S*)/$1/) {
      $_[KERNEL]->yield( do => sub {
        get_link_id( "web", @_[ARG0..ARG1], "[link]" );
      }, "$heap->{remote_addr}:$heap->{remote_port}", $1 );

      $response->content
        ( "<html><head><title>$heap->{my_name} thanks you</title></head>" .
          "<body>" .
          "<p>" .
          "Thanks for submitting <tt>$link</tt> to " .
          "<a href='/'>$heap->{my_name}</a>." .
          "</p>" .
          "</body>" .
          "</html>"
        );
    }
    else {
      $response->content
        ( "<html><head><title>$heap->{my_name} thanks you</title></head>" .
          "<body>" .
          "<p>" .
          "Thanks for submitting <tt>$link</tt> to " .
          "<a href='/'>$heap->{my_name}</a>.  However, it cannot accept " .
          "this type of link at this time." .
          "</p>" .
          "</body>" .
          "</html>"
        );
    }

    $heap->{wheel}->put( $response );
    return;
  }

  if ($url =~ /^\/post$/) {

    # spawn off a session to do the actual parsing
    $_[KERNEL]->yield( do => sub {
      my $content = $_[ARG0];
      $content =~ tr/+/ /;
      $content =~ s/%([a-fA-F0-9]{2})/pack("C", hex($1))/eg;
      $content =~ s/^\w+=//;
      my ($description, @links) = parse_link_from_message($content);
      $description = "(none)"
        unless defined $description and length $description;

      foreach my $link (@links) {
        next unless defined $link and length $link;
        # spawn off a session to do the actual link fetching
        $_[KERNEL]->yield( do => sub {
          get_link_id("web", @_[ARG0..ARG1], "[email]");
        }, $_[ARG1], $link );
      }
    }, $request->content, "$heap->{remote_addr}:$heap->{remote_port}" );

    my $response = HTTP::Response->new(200);
    $response->content(
      'Thanks!'
    );

    $heap->{wheel}->put( $response );
    return;
  }

  ### Do basic authentication if specified in config file.

  my $do_auth = (defined($heap->{login}) and defined($heap->{passwd}));
  my ($login, $password) = $request->authorization_basic() if $do_auth;

  unless ( ! $do_auth or
	   (defined($login) and ($login eq $heap->{login}) and
	    defined($password) and ($password eq $heap->{passwd}))
         )
  {
    my $response = new HTTP::Response(401);
    $response->push_header('WWW-Authenticate', 'Basic realm="memephage"');
    $response->push_header('Server', 'memephage/1.0');
    $response->push_header('Content-Type', 'text/html');
    $response->content
      ( '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN"> ' .
        '<html>' .
        '<head>' .
        '<title>WARNING: You are in a restricted area.</title>' .
        '<link rev="made" href="mailto:troc@netrus.net">' .
        '</head>' .
        '<body bgcolor="#C00000" text="#FFFFFF" link="#00FF00" ' .
        'vlink="#00FF00">' .
        '<table border=0 cellspacing=0 width="100%" height="100%">' .
        '<tr>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td bgcolor="#000000" width="6%"><font size="7">&#160;</font></td>' .
        '<td align=center colspan=7><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="6%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td bgcolor="#FFFF00" width="6%"><font size="7">&#160;</font></td>' .
        '<td align=center colspan=7><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="6%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td bgcolor="#000000" width="6%"><font size="7">&#160;</font></td>' .
        '<td align=center valign=middle colspan=7 rowspan=3>'  .
        '<tt><font size="7"><b>* WARNING *<br>'  .
        '<br>You are in a restricted area.</b></font></tt></td>' .
        '<td bgcolor="#000000" width="6%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td bgcolor="#FFFF00" width="6%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="6%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td bgcolor="#000000" width="6%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="6%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td bgcolor="#FFFF00" width="6%"><font size="7">&#160;</font></td>' .
        '<td align=center colspan=7><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="6%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td bgcolor="#000000" width="6%"><font size="7">&#160;</font></td>' .
        '<td align=center colspan=7><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="6%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#000000" width="11%"><font size="7">&#160;</font></td>' .
        '<td bgcolor="#FFFF00" width="11%"><font size="7">&#160;</font></td>' .
        '</tr>' .
        '<tr>' .
        '<td colspan=9 align=right>' .
        '<p>' .
        '<font size="1">' .
        '<a href="http://poe.perl.org/">Powered by POE</a>' .
        '</font>' .
        '</td>'  .
        '</tr>'  .
        '</table>' .
        '</body>' .
        '</html>'
      );
    $heap->{wheel}->put($response);
    return;
  }

  ### Link redirection.


  ### Default handler dumps everything it can about the request.

  my $response = HTTP::Response->new( 200 );
  $response->push_header( 'Content-type', 'text/html' );

  # Many of the headers dumped here are undef.  We turn off warnings
  # here so the program doesn't constantly squeal.

  local $^W = 0;

  $response->content
    ( "<html><head><title>test</title></head>" .
      "<body>Your request was strange" .
      &dump_content($request->content()) .

      "<p>Request as string=" . $request->as_string() . "</p>" .

      "</body></html>"
    );

  # A little debugging here.
  if (DUMP_REQUEST) {
    my $request_as_string = $request->as_string();
    warn unpack('H*', $request_as_string), "\n";
    warn "Request has CR.\n" if $request_as_string =~ /\x0D/;
    warn "Request has LF.\n" if $request_as_string =~ /\x0A/;
  }

  $heap->{wheel}->put( $response );
  return;
}

# Start the HTTPD server.

foreach my $server (get_names_by_type(WEBLOG_TYPE)) {
  my %conf = get_items_by_name($server);

  POE::Component::Server::TCP->new
    ( Port     => $conf{port},
      ( (defined $conf{iface})
        ? ( Address => $conf{iface} )
        : ()
      ),
      Acceptor =>
      sub {
        POE::Session->new
          ( _start    => \&httpd_session_started,
            got_flush => \&httpd_session_flushed,
            got_query => \&httpd_session_got_query,
            got_error => \&httpd_session_got_error,
            do => sub {goto &{splice(@_, ARG0, 1)} if ref($_[ARG0]) =~ /CODE/},

            # Note the use of ifname here in ARG6.  This gives the
            # responding session knowledge of its host name for
            # building HTML responses.  Most of the time it will be
            # identical to iface, but sometimes there may be a reverse
            # proxy, firewall, or NATD between the address we bind to
            # and the one people connect to.  In that case, ifname is
            # the address the outside world sees, and iface is the one
            # we've bound to.

            [ @_[ARG0..ARG2], $server,
              $conf{iface}, $conf{port}, $conf{ifname},
	      $conf{login}, $conf{passwd}, $conf{template_root},
            ],
          );
      },
    );
}

#------------------------------------------------------------------------------

1;
